import { Plugin } from "ckeditor5/src/core";
import { toWidget, toWidgetEditable } from "ckeditor5/src/widget";
import { Widget } from "ckeditor5/src/widget";
import InsertjwBoxCommand from "./insertjwboxcommand";

// cSpell:ignore jwbox insertjwboxcommand

/**
 * CKEditor 5 plugins do not work directly with the DOM. They are defined as
 * plugin-specific data models that are then converted to markup that
 * is inserted in the DOM.
 *
 * CKEditor 5 internally interacts with jwBox as this model:
 * <jwBox>
 *    <jwScript></jwScript>
 * </jwBox>
 *
 * Which is converted for the browser/user as this markup
 * <section class="jw-box">
 *   <script src=""></script>
 * </section>
 *
 * This file has the logic for defining the jwBox model, and for how it is
 * converted to standard DOM markup.
 */
export default class jwBoxEditing extends Plugin {
  static get requires() {
    return [Widget];
  }

  init() {
    this._defineSchema();
    this._defineConverters();
    this.editor.commands.add(
      "insertjwBox",
      new InsertjwBoxCommand(this.editor)
    );
  }

  /*
   * This registers the structure that will be seen by CKEditor 5 as
   * <jwBox>
   *    <jwScript></jwScript>
   * </jwBox>
   *
   * The logic in _defineConverters() will determine how this is converted to
   * markup.
   */
  _defineSchema() {
    // Schemas are registered via the central `editor` object.
    const schema = this.editor.model.schema;

    schema.register("jwBox", {
      // Behaves like a self-contained object (e.g. an image).
      isObject: true,
      // Allow in places where other blocks are allowed (e.g. directly in the root).
      allowWhere: "$block",
    });

    schema.register("jwScript", {
      // Cannot be split or left by the caret.
      isObject: true,
      isLimit: true,
      allowIn: "jwBox",
      // the actual src
      allowAttributes: ["src"],
      src: {
        // Allow any value.
        allowAll: true,
      },
    });
  }

  /**
   * Converters determine how CKEditor 5 models are converted into markup and
   * vice-versa.
   */
  _defineConverters() {
    // Converters are registered via the central editor object.
    const { conversion } = this.editor;

    // Upcast Converters: determine how existing HTML is interpreted by the
    // editor. These trigger when an editor instance loads.
    //
    // If <section class="jwbox"> is present in the existing markup
    // processed by CKEditor, then CKEditor recognizes and loads it as a
    // <jwBox> model.
    conversion.for("upcast").elementToElement({
      model: "jwBox",
      view: {
        name: "section",
        classes: "jw-box",
      },
    });

    // <jwScript> model.
    conversion.for("upcast").elementToElement({
      model: "jwScript",
      view: {
        name: "script",
      },
    }).attributeToAttribute({
      view: {
        name: 'script',
        key: 'src',
      },
      model: {
        key: 'src',
        value: (viewElement) => {

            return viewElement.getAttribute('src');
        },
      },
    });
    // <section class="jw-box">{{inner content}}</section>.
    conversion.for("dataDowncast").elementToElement({
      model: "jwBox",
      view: {
        name: "section",
        classes: "jw-box",
      },
    });
    conversion.for("dataDowncast").elementToElement({
      model: "jwScript",
      view: (modelElement, { writer: viewWriter }) => {
        const script = viewWriter.createContainerElement("script", {
          src: modelElement.getAttribute("src"),
        }, {
          isAllowedInsideAttributeElement: true
        });
        return toWidget(script, viewWriter);
      },
    });
    // Convert the <jwBox> model into a container widget in the editor UI.
    conversion.for("editingDowncast").elementToElement({
      model: "jwBox",
      view: (modelElement, { writer: viewWriter }) => {
        const section = viewWriter.createContainerElement("section", {
          class: "jw-box",
        });

        return toWidget(section, viewWriter, { label: "simple box widget" });
      },
    });
    // Convert the <jwScript> model into a container widget in the editor UI.
    conversion.for("editingDowncast").elementToElement({
      model: "jwScript",
      view: (modelElement, { writer: viewWriter }) => {
        const script = viewWriter.createContainerElement("script", {
          src: modelElement.getAttribute("src"),
        }, {
          isAllowedInsideAttributeElement: true
        });

        return toWidget(script, viewWriter);
      },
    });

  }
}
